#include <stdlib.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/interrupt.h>

enum {
	NOP = 0,
	RESET,
	SPEED,
	SET_LED,
	SET_LED_RANDOM,
	SET_LEDS,
	SET_LEDS_RANDOM,
	SET_ALL,
	UNSET_LED,
	UNSET_ALL,
	ROTATE_CLOCKWISE,
	ROTATE_ANTI_CLOCKWISE,
	ROTATE_OUT,
	ROTATE_IN,
	FOR,
	LOOP
};

#define LEDS 19
unsigned char ledMappings[] PROGMEM = {
	0x02, // centre
	0x32, 0x23, 0x31, 0x13, 0x41, 0x14, // inner ring
	0x34, 0x24, 0x12, 0x01, 0x30, 0x40, // middle ring
	0x43, 0x42, 0x21, 0x10, 0x03, 0x04, // outer ring
	0x20 // unused!
};

/*
#define FRAMES 26
unsigned long frames[] PROGMEM = {
	0x00000001, 0x0000007e, 0x00001f80, 0x000fe000, // centre-out
	0x00001f80, 0x0000007e, // outside-to-centre
	0x00002083, 0x00004105, 0x00008209, 0x00010411, 0x00020821, 0x00041041, 0x00002083, // anti-clockwise rotation
	0x00000001, 0x0000007e, 0x00001f80, 0x000fe000, // centre-out
	0x00001f80, 0x0000007e, // outside-to-centre
	0x00002083, 0x00041041, 0x00020821, 0x00010411, 0x00008209, 0x00004105, 0x00002083 // clockwise rotation
};
*/

unsigned char program[] PROGMEM = {
	SPEED, 2,
	FOR, 40,
		UNSET_ALL,
		SET_LEDS_RANDOM,
		NOP,
	LOOP,

	SPEED, 1,
	FOR, 100,
		UNSET_ALL,
		SET_LED_RANDOM,
		NOP,
	LOOP,

	SPEED, 8,
	SET_LEDS, 0x01, 0, 0, 0,
	FOR, 3,
		ROTATE_OUT,
		NOP,
	LOOP,
	FOR, 3,
		ROTATE_IN,
		NOP,
	LOOP,

	SET_LEDS, 0x01, 0, 0, 0,
	FOR, 3,
		ROTATE_OUT,
		SET_LED, 0,
		NOP,
	LOOP,
	FOR, 3,
		UNSET_LED, 0,
		ROTATE_IN,
		NOP,
	LOOP,
	UNSET_LED, 0,
	NOP,

	SPEED, 4,
	SET_LEDS, 0x93, 0x20, 0, 0,
	FOR, 12,
		ROTATE_ANTI_CLOCKWISE,
		NOP,
	LOOP,
	FOR, 12,
		ROTATE_CLOCKWISE,
		NOP,
	LOOP,

	SPEED, 8,
	UNSET_ALL,
	FOR, 5,
		SET_LED_RANDOM,
	LOOP,
	NOP,
	FOR, 12,
		ROTATE_ANTI_CLOCKWISE,
		NOP,
	LOOP,
	FOR, 12,
		ROTATE_CLOCKWISE,
		NOP,
	LOOP,

	UNSET_ALL,
	RESET
};

unsigned long displayState;
unsigned char speed;
unsigned char interrupts;
unsigned char instruction;
unsigned char step;
unsigned char stack;
unsigned char counter;

ISR(TIM0_COMPA_vect) {
	interrupts++;
	if(interrupts == speed) {
		interrupts = 0;
		step = 1;
	}
}

void interpret() {
	step = 0;
	switch(pgm_read_byte(&program[instruction++])) {
		case NOP:
			// nothing, hold display state
			break;
		case RESET:
			instruction = 0;
			step = 1;
			break;
		case SPEED:
			speed = pgm_read_byte(&program[instruction++]);
			step = 1;
			break;
		case UNSET_ALL:
			displayState = 0;
			step = 1;
			break;
		case SET_ALL:
			displayState = 0x0007FFFF;
			step = 1;
			break;
		case SET_LED:
			displayState |= 1L << pgm_read_byte(&program[instruction++]);
			step = 1;
			break;
		case UNSET_LED:
			displayState &= ~(1L << pgm_read_byte(&program[instruction++]));
			step = 1;
			break;
		case SET_LED_RANDOM:
			displayState |= 1 << (random() % 18);
			step = 1;
			break;
		case SET_LEDS_RANDOM:
			displayState = random();
			break;
		case SET_LEDS:
			displayState = pgm_read_dword(&program[instruction]);
			instruction += 4;
			break;
		case ROTATE_CLOCKWISE:
			displayState =  (displayState & 0x1L) | ((displayState & 0x7DF7CL) >> 1) | ((displayState & 0x2082L) << 5);
			step = 1;
			break;
		case ROTATE_ANTI_CLOCKWISE:
			displayState =  (displayState & 0x1L) | ((displayState & 0x3EFBEL) << 1) | ((displayState & 0x41040L) >> 5);
			step = 1;
			break;
		case ROTATE_OUT:
			displayState =  ((displayState & 0x1L)?(0x7EL):(0)) | ((displayState & 0x1FFEL) << 6) | ((displayState & 0x7E000L)?(0x1L):(0));
			step = 1;
			break;
		case ROTATE_IN:
			displayState =  ((displayState & 0x1L)?(0x7E000L):(0)) | ((displayState & 0x7FF80) >> 6) | ((displayState & 0x7EL)?(0x1L):(0));
			step = 1;
			break;
		case FOR:
			counter = pgm_read_byte(&program[instruction++]);
			stack = instruction;
			step = 1;
			break;
		case LOOP:
			if(--counter) instruction = stack;
			step = 1;
			break;
		default:
			// skip unknown instructions
			step = 1;
			break;
	}
}

void charlieplex() {
	unsigned char led;
	unsigned char map, set, unset;

	for(led = 0; led < LEDS; led++) {
		map = pgm_read_byte(&ledMappings[led]);
		set = 1 << (map & 0x0f);
		unset = 1 << ((map & 0xf0) >> 4);
		PORTB = 0;
		DDRB = set | unset;
		if(displayState & (1L << led)) PORTB = set;
		_delay_us(200);
	}

	DDRB = 0;
	PORTB = 0;
}

void init() {
	step = 0;
	speed = 8;
	instruction = 0;
	displayState = 0;
	
	MCUCR |= _BV(PUD);		// disable pull-ups globally

	TCCR0A = _BV(WGM01);		// Clear timer on compare match mode
	TCCR0B = _BV(CS02) | _BV(CS00);	// Clock prescale by 1024
	OCR0A = 0xff;			// Compare match at top
	TIMSK = _BV(OCIE0A);		// Enable compare match interrupts

	sei();
}

int main() {

	init();
	while(1) {
		while(step) interpret();
		charlieplex();
	}

	return 0; // unreached
}
